from __future__ import annotations

from .client import (
    BoundStorageBox,
    BoundStorageBoxSnapshot,
    BoundStorageBoxSubaccount,
    StorageBoxesClient,
    StorageBoxesPageResult,
    StorageBoxSnapshotsPageResult,
    StorageBoxSubaccountsPageResult,
)
from .domain import (
    CreateStorageBoxResponse,
    CreateStorageBoxSnapshotResponse,
    CreateStorageBoxSubaccountResponse,
    DeleteStorageBoxResponse,
    DeleteStorageBoxSnapshotResponse,
    DeleteStorageBoxSubaccountResponse,
    StorageBox,
    StorageBoxAccessSettings,
    StorageBoxFoldersResponse,
    StorageBoxSnapshot,
    StorageBoxSnapshotPlan,
    StorageBoxSnapshotStats,
    StorageBoxStats,
    StorageBoxStatus,
    StorageBoxSubaccount,
    StorageBoxSubaccountAccessSettings,
)

__all__ = [
    "BoundStorageBox",
    "BoundStorageBoxSnapshot",
    "BoundStorageBoxSubaccount",
    "CreateStorageBoxResponse",
    "CreateStorageBoxSnapshotResponse",
    "CreateStorageBoxSubaccountResponse",
    "DeleteStorageBoxResponse",
    "DeleteStorageBoxSnapshotResponse",
    "DeleteStorageBoxSubaccountResponse",
    "StorageBox",
    "StorageBoxAccessSettings",
    "StorageBoxesClient",
    "StorageBoxesPageResult",
    "StorageBoxFoldersResponse",
    "StorageBoxSnapshot",
    "StorageBoxSnapshotPlan",
    "StorageBoxSnapshotsPageResult",
    "StorageBoxSnapshotStats",
    "StorageBoxStats",
    "StorageBoxStatus",
    "StorageBoxSubaccount",
    "StorageBoxSubaccountAccessSettings",
    "StorageBoxSubaccountsPageResult",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from ..actions import BoundAction
from ..core import BaseDomain, DomainIdentityMixin
from ..locations import BoundLocation, Location
from ..storage_box_types import BoundStorageBoxType, StorageBoxType

if TYPE_CHECKING:
    from .client import (
        BoundStorageBox,
        BoundStorageBoxSnapshot,
        BoundStorageBoxSubaccount,
    )

__all__ = [
    "StorageBox",
    "StorageBoxAccessSettings",
    "StorageBoxStats",
    "StorageBoxSnapshotPlan",
    "CreateStorageBoxResponse",
    "DeleteStorageBoxResponse",
    "StorageBoxFoldersResponse",
    "StorageBoxSnapshot",
    "StorageBoxSnapshotStats",
    "CreateStorageBoxSnapshotResponse",
    "DeleteStorageBoxSnapshotResponse",
    "StorageBoxSubaccount",
    "StorageBoxSubaccountAccessSettings",
    "CreateStorageBoxSubaccountResponse",
    "DeleteStorageBoxSubaccountResponse",
    "StorageBoxStatus",
]


StorageBoxStatus = Literal[
    "active",
    "initializing",
    "locked",
]


class StorageBox(BaseDomain, DomainIdentityMixin):
    """
    Storage Box Domain.

    See https://docs.hetzner.cloud/reference/hetzner#storage-boxes.
    """

    STATUS_ACTIVE = "active"
    STATUS_INITIALIZING = "initializing"
    STATUS_LOCKED = "locked"

    __api_properties__ = (
        "id",
        "name",
        "storage_box_type",
        "location",
        "system",
        "server",
        "username",
        "labels",
        "protection",
        "snapshot_plan",
        "access_settings",
        "stats",
        "status",
        "created",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        id: int | None = None,
        name: str | None = None,
        storage_box_type: BoundStorageBoxType | StorageBoxType | None = None,
        location: BoundLocation | Location | None = None,
        system: str | None = None,
        server: str | None = None,
        username: str | None = None,
        labels: dict[str, str] | None = None,
        protection: dict[str, bool] | None = None,
        snapshot_plan: StorageBoxSnapshotPlan | None = None,
        access_settings: StorageBoxAccessSettings | None = None,
        stats: StorageBoxStats | None = None,
        status: StorageBoxStatus | None = None,
        created: str | None = None,
    ):
        self.id = id
        self.name = name
        self.storage_box_type = storage_box_type
        self.location = location
        self.system = system
        self.server = server
        self.username = username
        self.labels = labels
        self.protection = protection
        self.snapshot_plan = snapshot_plan
        self.access_settings = access_settings
        self.stats = stats
        self.status = status
        self.created = self._parse_datetime(created)


class StorageBoxAccessSettings(BaseDomain):
    """
    Storage Box Access Settings Domain.
    """

    __api_properties__ = (
        "reachable_externally",
        "samba_enabled",
        "ssh_enabled",
        "webdav_enabled",
        "zfs_enabled",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        reachable_externally: bool | None = None,
        samba_enabled: bool | None = None,
        ssh_enabled: bool | None = None,
        webdav_enabled: bool | None = None,
        zfs_enabled: bool | None = None,
    ):
        self.reachable_externally = reachable_externally
        self.samba_enabled = samba_enabled
        self.ssh_enabled = ssh_enabled
        self.webdav_enabled = webdav_enabled
        self.zfs_enabled = zfs_enabled

    def to_payload(self) -> dict[str, Any]:
        """
        Generates the request payload from this domain object.
        """
        payload: dict[str, Any] = {}
        if self.reachable_externally is not None:
            payload["reachable_externally"] = self.reachable_externally
        if self.samba_enabled is not None:
            payload["samba_enabled"] = self.samba_enabled
        if self.ssh_enabled is not None:
            payload["ssh_enabled"] = self.ssh_enabled
        if self.webdav_enabled is not None:
            payload["webdav_enabled"] = self.webdav_enabled
        if self.zfs_enabled is not None:
            payload["zfs_enabled"] = self.zfs_enabled
        return payload


class StorageBoxStats(BaseDomain):
    """
    Storage Box Stats Domain.
    """

    __api_properties__ = (
        "size",
        "size_data",
        "size_snapshots",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        size: int | None = None,
        size_data: int | None = None,
        size_snapshots: int | None = None,
    ):
        self.size = size
        self.size_data = size_data
        self.size_snapshots = size_snapshots


class StorageBoxSnapshotPlan(BaseDomain):
    """
    Storage Box Snapshot Plan Domain.
    """

    __api_properties__ = (
        "max_snapshots",
        "hour",
        "minute",
        "day_of_week",
        "day_of_month",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        max_snapshots: int,
        hour: int,
        minute: int,
        day_of_week: int | None = None,
        day_of_month: int | None = None,
    ):
        self.max_snapshots = max_snapshots
        self.hour = hour
        self.minute = minute
        self.day_of_week = day_of_week
        self.day_of_month = day_of_month

    def to_payload(self) -> dict[str, Any]:
        """
        Generates the request payload from this domain object.
        """
        payload: dict[str, Any] = {
            "max_snapshots": self.max_snapshots,
            "hour": self.hour,
            "minute": self.minute,
            "day_of_week": self.day_of_week,  # API default is null
            "day_of_month": self.day_of_month,  # API default is null
        }

        return payload


class CreateStorageBoxResponse(BaseDomain):
    """
    Create Storage Box Response Domain.
    """

    __api_properties__ = (
        "storage_box",
        "action",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        storage_box: BoundStorageBox,
        action: BoundAction,
    ):
        self.storage_box = storage_box
        self.action = action


class DeleteStorageBoxResponse(BaseDomain):
    """
    Delete Storage Box Response Domain.
    """

    __api_properties__ = ("action",)
    __slots__ = __api_properties__

    def __init__(
        self,
        action: BoundAction,
    ):
        self.action = action


class StorageBoxFoldersResponse(BaseDomain):
    """
    Storage Box Folders Response Domain.
    """

    __api_properties__ = ("folders",)
    __slots__ = __api_properties__

    def __init__(
        self,
        folders: list[str],
    ):
        self.folders = folders


# Snapshots
###############################################################################


class StorageBoxSnapshot(BaseDomain, DomainIdentityMixin):
    """
    Storage Box Snapshot Domain.
    """

    __api_properties__ = (
        "id",
        "name",
        "description",
        "is_automatic",
        "labels",
        "storage_box",
        "created",
        "stats",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        id: int | None = None,
        name: str | None = None,
        description: str | None = None,
        is_automatic: bool | None = None,
        labels: dict[str, str] | None = None,
        storage_box: BoundStorageBox | StorageBox | None = None,
        created: str | None = None,
        stats: StorageBoxSnapshotStats | None = None,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.is_automatic = is_automatic
        self.labels = labels
        self.storage_box = storage_box
        self.created = self._parse_datetime(created)
        self.stats = stats


class StorageBoxSnapshotStats(BaseDomain):
    """
    Storage Box Snapshot Stats Domain.
    """

    __api_properties__ = (
        "size",
        "size_filesystem",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        size: int,
        size_filesystem: int,
    ):
        self.size = size
        self.size_filesystem = size_filesystem


class CreateStorageBoxSnapshotResponse(BaseDomain):
    """
    Create Storage Box Snapshot Response Domain.
    """

    __api_properties__ = (
        "snapshot",
        "action",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        snapshot: BoundStorageBoxSnapshot,
        action: BoundAction,
    ):
        self.snapshot = snapshot
        self.action = action


class DeleteStorageBoxSnapshotResponse(BaseDomain):
    """
    Delete Storage Box Snapshot Response Domain.
    """

    __api_properties__ = ("action",)
    __slots__ = __api_properties__

    def __init__(
        self,
        action: BoundAction,
    ):
        self.action = action


# Subaccounts
###############################################################################


class StorageBoxSubaccount(BaseDomain, DomainIdentityMixin):
    """
    Storage Box Subaccount Domain.
    """

    __api_properties__ = (
        "id",
        "name",
        "username",
        "description",
        "server",
        "home_directory",
        "access_settings",
        "labels",
        "storage_box",
        "created",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        id: int | None = None,
        name: str | None = None,
        username: str | None = None,
        description: str | None = None,
        server: str | None = None,
        home_directory: str | None = None,
        access_settings: StorageBoxSubaccountAccessSettings | None = None,
        labels: dict[str, str] | None = None,
        storage_box: BoundStorageBox | StorageBox | None = None,
        created: str | None = None,
    ):
        self.id = id
        self.name = name
        self.username = username
        self.description = description
        self.server = server
        self.home_directory = home_directory
        self.access_settings = access_settings
        self.labels = labels
        self.storage_box = storage_box
        self.created = self._parse_datetime(created)


class StorageBoxSubaccountAccessSettings(BaseDomain):
    """
    Storage Box Subaccount Access Settings Domain.
    """

    __api_properties__ = (
        "reachable_externally",
        "samba_enabled",
        "ssh_enabled",
        "webdav_enabled",
        "readonly",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        reachable_externally: bool | None = None,
        samba_enabled: bool | None = None,
        ssh_enabled: bool | None = None,
        webdav_enabled: bool | None = None,
        readonly: bool | None = None,
    ):
        self.reachable_externally = reachable_externally
        self.samba_enabled = samba_enabled
        self.ssh_enabled = ssh_enabled
        self.webdav_enabled = webdav_enabled
        self.readonly = readonly

    def to_payload(self) -> dict[str, Any]:
        """
        Generates the request payload from this domain object.
        """
        payload: dict[str, Any] = {}
        if self.reachable_externally is not None:
            payload["reachable_externally"] = self.reachable_externally
        if self.samba_enabled is not None:
            payload["samba_enabled"] = self.samba_enabled
        if self.ssh_enabled is not None:
            payload["ssh_enabled"] = self.ssh_enabled
        if self.webdav_enabled is not None:
            payload["webdav_enabled"] = self.webdav_enabled
        if self.readonly is not None:
            payload["readonly"] = self.readonly
        return payload


class CreateStorageBoxSubaccountResponse(BaseDomain):
    """
    Create Storage Box Subaccount Response Domain.
    """

    __api_properties__ = (
        "subaccount",
        "action",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        subaccount: BoundStorageBoxSubaccount,
        action: BoundAction,
    ):
        self.subaccount = subaccount
        self.action = action


class DeleteStorageBoxSubaccountResponse(BaseDomain):
    """
    Delete Storage Box Subaccount Response Domain.
    """

    __api_properties__ = ("action",)
    __slots__ = __api_properties__

    def __init__(
        self,
        action: BoundAction,
    ):
        self.action = action



from __future__ import annotations

from typing import TYPE_CHECKING, Any, NamedTuple

from ..actions import (
    ActionSort,
    ActionsPageResult,
    ActionStatus,
    BoundAction,
    ResourceActionsClient,
)
from ..actions.client import ResourceClientBaseActionsMixin
from ..core import BoundModelBase, Meta, ResourceClientBase
from .domain import (
    CreateZoneResponse,
    CreateZoneRRSetResponse,
    DeleteZoneResponse,
    DeleteZoneRRSetResponse,
    ExportZonefileResponse,
    Zone,
    ZoneAuthoritativeNameservers,
    ZoneMode,
    ZonePrimaryNameserver,
    ZoneRecord,
    ZoneRRSet,
    ZoneRRSetType,
)

if TYPE_CHECKING:
    from .._client import Client

__all__ = [
    "BoundZone",
    "BoundZoneRRSet",
    "ZonesPageResult",
    "ZoneRRSetsPageResult",
    "ZonesClient",
]


class BoundZone(BoundModelBase[Zone], Zone):
    _client: ZonesClient

    model = Zone

    def __init__(
        self,
        client: ZonesClient,
        data: dict[str, Any],
        complete: bool = True,
    ):
        raw = data.get("primary_nameservers")
        if raw is not None:
            data["primary_nameservers"] = [
                ZonePrimaryNameserver.from_dict(o) for o in raw
            ]

        raw = data.get("authoritative_nameservers")
        if raw:
            data["authoritative_nameservers"] = ZoneAuthoritativeNameservers.from_dict(
                raw
            )

        super().__init__(client, data, complete)

    def update(
        self,
        *,
        labels: dict[str, str] | None = None,
    ) -> BoundZone:
        """
        Updates the Zone.

        See https://docs.hetzner.cloud/reference/cloud#zones-update-a-zone

        :param labels: User-defined labels (key/value pairs) for the Resource.
        """
        return self._client.update(self, labels=labels)

    def delete(self) -> DeleteZoneResponse:
        """
        Deletes the Zone.

        See https://docs.hetzner.cloud/reference/cloud#zones-delete-a-zone
        """
        return self._client.delete(self)

    def export_zonefile(self) -> ExportZonefileResponse:
        """
        Returns a generated Zone file in BIND (RFC 1034/1035) format.

        See https://docs.hetzner.cloud/reference/cloud#zones-export-a-zone-file
        """
        return self._client.export_zonefile(self)

    def get_actions_list(
        self,
        *,
        status: list[ActionStatus] | None = None,
        sort: list[ActionSort] | None = None,
        page: int | None = None,
        per_page: int | None = None,
    ) -> ActionsPageResult:
        """
        Returns a paginated list of Actions for a Zone.

        See https://docs.hetzner.cloud/reference/cloud#zones-list-zones

        :param status: Filter the Actions by status.
        :param sort: Sort Actions by field and direction.
        :param page: Page number to get.
        :param per_page: Maximum number of Actions returned per page.
        """
        return self._client.get_actions_list(
            self,
            status=status,
            sort=sort,
            page=page,
            per_page=per_page,
        )

    def get_actions(
        self,
        *,
        status: list[ActionStatus] | None = None,
        sort: list[ActionSort] | None = None,
    ) -> list[BoundAction]:
        """
        Returns all Actions for a Zone.

        See https://docs.hetzner.cloud/reference/cloud#zones-list-zones

        :param status: Filter the Actions by status.
        :param sort: Sort Actions by field and direction.
        """
        return self._client.get_actions(
            self,
            status=status,
            sort=sort,
        )

    def import_zonefile(
        self,
        zonefile: str,
    ) -> BoundAction:
        """
        Imports a zone file, replacing all resource record sets (ZoneRRSet).

        See https://docs.hetzner.cloud/reference/cloud#zone-actions-import-a-zone-file

        :param zonefile: Zone file to import.
        """
        return self._client.import_zonefile(self, zonefile=zonefile)

    def change_protection(
        self,
        *,
        delete: bool | None = None,
    ) -> BoundAction:
        """
        Changes the protection of the Zone.

        See https://docs.hetzner.cloud/reference/cloud#zone-actions-change-a-zones-protection

        :param delete: Prevents the Zone from being deleted.
        """
        return self._client.change_protection(self, delete=delete)

    def change_ttl(
        self,
        ttl: int,
    ) -> BoundAction:
        """
        Changes the TTL of the Zone.

        See https://docs.hetzner.cloud/reference/cloud#zone-actions-change-a-zones-default-ttl

        :param ttl: Default Time To Live (TTL) of the Zone.
        """
        return self._client.change_ttl(self, ttl=ttl)

    def change_primary_nameservers(
        self,
        primary_nameservers: list[ZonePrimaryNameserver],
    ) -> BoundAction:
        """
        Changes the primary nameservers of the Zone.

        See https://docs.hetzner.cloud/reference/cloud#zone-actions-change-a-zones-primary-nameservers

        :param primary_nameservers: Primary nameservers of the Zone.
        """
        return self._client.change_primary_nameservers(
            self,
            primary_nameservers=primary_nameservers,
        )

    def get_rrset(
        self,
        name: str,
        type: ZoneRRSetType,
    ) -> BoundZoneRRSet:
        """
        Returns a single ZoneRRSet from the Zone.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-get-an-rrset

        :param name: Name of the RRSet.
        :param type: Type of the RRSet.
        """
        return self._client.get_rrset(self, name=name, type=type)

    def get_rrset_list(
        self,
        *,
        name: str | None = None,
        type: list[ZoneRRSetType] | None = None,
        label_selector: str | None = None,
        sort: list[str] | None = None,
        page: int | None = None,
        per_page: int | None = None,
    ) -> ZoneRRSetsPageResult:
        """
        Returns all ZoneRRSet in the Zone for a specific page.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-list-rrsets

        :param name: Filter resources by their name. The response will only contain the resources matching exactly the specified name.
        :param type: Filter resources by their type. The response will only contain the resources matching exactly the specified type.
        :param label_selector: Filter resources by labels. The response will only contain resources matching the label selector.
        :param sort: Sort resources by field and direction.
        :param page: Page number to return.
        :param per_page: Maximum number of entries returned per page.
        """
        return self._client.get_rrset_list(
            self,
            name=name,
            type=type,
            label_selector=label_selector,
            sort=sort,
            page=page,
            per_page=per_page,
        )

    def get_rrset_all(
        self,
        *,
        name: str | None = None,
        type: list[ZoneRRSetType] | None = None,
        label_selector: str | None = None,
        sort: list[str] | None = None,
    ) -> list[BoundZoneRRSet]:
        """
        Returns all ZoneRRSet in the Zone.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-list-rrsets

        :param name: Filter resources by their name. The response will only contain the resources matching exactly the specified name.
        :param type: Filter resources by their type. The response will only contain the resources matching exactly the specified type.
        :param label_selector: Filter resources by labels. The response will only contain resources matching the label selector.
        :param sort: Sort resources by field and direction.
        """
        return self._client.get_rrset_all(
            self,
            name=name,
            type=type,
            label_selector=label_selector,
            sort=sort,
        )

    def create_rrset(
        self,
        *,
        name: str,
        type: ZoneRRSetType,
        ttl: int | None = None,
        labels: dict[str, str] | None = None,
        records: list[ZoneRecord] | None = None,
    ) -> CreateZoneRRSetResponse:
        """
        Creates a ZoneRRSet in the Zone.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-create-an-rrset

        :param name: Name of the RRSet.
        :param type: Type of the RRSet.
        :param ttl: Time To Live (TTL) of the RRSet.
        :param labels: User-defined labels (key/value pairs) for the Resource.
        :param records: Records of the RRSet.
        """
        return self._client.create_rrset(
            self,
            name=name,
            type=type,
            ttl=ttl,
            labels=labels,
            records=records,
        )

    def update_rrset(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        *,
        labels: dict[str, str] | None = None,
    ) -> BoundZoneRRSet:
        """
        Updates a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-update-an-rrset

        :param rrset: RRSet to update.
        :param labels: User-defined labels (key/value pairs) for the Resource.
        """
        return self._client.update_rrset(rrset=rrset, labels=labels)

    def delete_rrset(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
    ) -> DeleteZoneRRSetResponse:
        """
        Deletes a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-delete-an-rrset

        :param rrset: RRSet to delete.
        """
        return self._client.delete_rrset(rrset=rrset)

    def change_rrset_protection(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        *,
        change: bool | None = None,
    ) -> BoundAction:
        """
        Changes the protection of a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-change-an-rrsets-protection

        :param rrset: RRSet to update.
        :param change: Prevent the Zone from being changed (deletion and updates).
        """
        return self._client.change_rrset_protection(rrset=rrset, change=change)

    def change_rrset_ttl(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        ttl: int | None,
    ) -> BoundAction:
        """
        Changes the TTL of a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-change-an-rrsets-ttl

        :param rrset: RRSet to update.
        :param change: Time To Live (TTL) of the RRSet.
        """
        return self._client.change_rrset_ttl(rrset=rrset, ttl=ttl)

    def add_rrset_records(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        records: list[ZoneRecord],
        ttl: int | None = None,
    ) -> BoundAction:
        """
        Adds records to a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-add-records-to-an-rrset

        :param rrset: RRSet to update.
        :param records: Records to add to the RRSet.
        :param ttl: Time To Live (TTL) of the RRSet.
        """
        return self._client.add_rrset_records(rrset=rrset, records=records, ttl=ttl)

    def update_rrset_records(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        records: list[ZoneRecord],
    ) -> BoundAction:
        """
        Updates records in a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-update-records-of-an-rrset

        :param rrset: RRSet to update.
        :param records: Records to update in the RRSet.
        """
        return self._client.update_rrset_records(rrset=rrset, records=records)

    def remove_rrset_records(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        records: list[ZoneRecord],
    ) -> BoundAction:
        """
        Removes records from a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-remove-records-from-an-rrset

        :param rrset: RRSet to update.
        :param records: Records to remove from the RRSet.
        """
        return self._client.remove_rrset_records(rrset=rrset, records=records)

    def set_rrset_records(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        records: list[ZoneRecord],
    ) -> BoundAction:
        """
        Sets the records of a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-set-records-of-an-rrset

        :param rrset: RRSet to update.
        :param records: Records to set in the RRSet.
        """
        return self._client.set_rrset_records(rrset=rrset, records=records)


class BoundZoneRRSet(BoundModelBase[ZoneRRSet], ZoneRRSet):
    _client: ZonesClient

    model = ZoneRRSet

    def __init__(
        self,
        client: ZonesClient,
        data: dict[str, Any],
        complete: bool = True,
    ):
        raw = data.get("zone")
        if raw is not None:
            data["zone"] = BoundZone(client, data={"id": raw}, complete=False)

        raw = data.get("records")
        if raw is not None:
            data["records"] = [ZoneRecord.from_dict(o) for o in raw]

        super().__init__(client, data, complete)

    def _get_self(self) -> BoundZoneRRSet:
        assert self.data_model.zone is not None
        assert self.data_model.type is not None
        return self._client.get_rrset(
            self.data_model.zone,
            self.data_model.name,
            self.data_model.type,
        )

    def update_rrset(
        self,
        *,
        labels: dict[str, str] | None = None,
    ) -> BoundZoneRRSet:
        """
        Updates the ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-update-an-rrset

        :param labels: User-defined labels (key/value pairs) for the Resource.
        """
        return self._client.update_rrset(self, labels=labels)

    def delete_rrset(
        self,
    ) -> DeleteZoneRRSetResponse:
        """
        Deletes the ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-delete-an-rrset
        """
        return self._client.delete_rrset(self)

    def change_rrset_protection(
        self,
        *,
        change: bool | None = None,
    ) -> BoundAction:
        """
        Changes the protection of the ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-change-an-rrsets-protection

        :param change: Prevent the Zone from being changed (deletion and updates).
        """
        return self._client.change_rrset_protection(self, change=change)

    def change_rrset_ttl(
        self,
        ttl: int | None,
    ) -> BoundAction:
        """
        Changes the TTL of the ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-change-an-rrsets-ttl

        :param change: Time To Live (TTL) of the RRSet.
        """
        return self._client.change_rrset_ttl(self, ttl=ttl)

    def add_rrset_records(
        self,
        records: list[ZoneRecord],
        ttl: int | None = None,
    ) -> BoundAction:
        """
        Adds records to the ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-add-records-to-an-rrset

        :param records: Records to add to the RRSet.
        :param ttl: Time To Live (TTL) of the RRSet.
        """
        return self._client.add_rrset_records(self, records=records, ttl=ttl)

    def update_rrset_records(
        self,
        records: list[ZoneRecord],
    ) -> BoundAction:
        """
        Updates records in a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-update-records-of-an-rrset

        :param records: Records to update in the RRSet.
        """
        return self._client.update_rrset_records(self, records=records)

    def remove_rrset_records(
        self,
        records: list[ZoneRecord],
    ) -> BoundAction:
        """
        Removes records from the ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-remove-records-from-an-rrset

        :param records: Records to remove from the RRSet.
        """
        return self._client.remove_rrset_records(self, records=records)

    def set_rrset_records(
        self,
        records: list[ZoneRecord],
    ) -> BoundAction:
        """
        Sets the records of the ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-set-records-of-an-rrset

        :param records: Records to set in the RRSet.
        """
        return self._client.set_rrset_records(self, records=records)


class ZonesPageResult(NamedTuple):
    zones: list[BoundZone]
    meta: Meta


class ZoneRRSetsPageResult(NamedTuple):
    rrsets: list[BoundZoneRRSet]
    meta: Meta


class ZonesClient(
    ResourceClientBaseActionsMixin,
    ResourceClientBase,
):
    """
    ZoneClient is a client for the Zone (DNS) API.

    See https://docs.hetzner.cloud/reference/cloud#zones and https://docs.hetzner.cloud/reference/cloud#zone-rrsets.
    """

    _base_url = "/zones"

    actions: ResourceActionsClient
    """Zones scoped actions client

    :type: :class:`ResourceActionsClient <hcloud.actions.client.ResourceActionsClient>`
    """

    def __init__(self, client: Client):
        super().__init__(client)
        self.actions = ResourceActionsClient(client, self._base_url)

    def get(self, id_or_name: int | str) -> BoundZone:
        """
        Returns a single Zone.

        See https://docs.hetzner.cloud/reference/cloud#zones-get-a-zone

        :param id_or_name: ID or Name of the Zone.
        """
        response = self._client.request(
            method="GET",
            url=f"{self._base_url}/{id_or_name}",
        )
        return BoundZone(self, response["zone"])

    def get_list(
        self,
        *,
        name: str | None = None,
        mode: ZoneMode | None = None,
        label_selector: str | None = None,
        sort: list[str] | None = None,
        page: int | None = None,
        per_page: int | None = None,
    ) -> ZonesPageResult:
        """
        Returns a list of Zone for a specific page.

        See https://docs.hetzner.cloud/reference/cloud#zones-list-zones

        :param name: Filter resources by their name. The response will only contain the resources matching exactly the specified name.
        :param mode: Filter resources by their mode. The response will only contain the resources matching exactly the specified mode.
        :param label_selector: Filter resources by labels. The response will only contain resources matching the label selector.
        :param sort: Sort resources by field and direction.
        :param page: Page number to return.
        :param per_page: Maximum number of entries returned per page.
        """
        params: dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        if mode is not None:
            params["mode"] = mode
        if label_selector is not None:
            params["label_selector"] = label_selector
        if sort is not None:
            params["sort"] = sort
        if page is not None:
            params["page"] = page
        if per_page is not None:
            params["per_page"] = per_page

        response = self._client.request(
            method="GET",
            url=f"{self._base_url}",
            params=params,
        )
        return ZonesPageResult(
            zones=[BoundZone(self, item) for item in response["zones"]],
            meta=Meta.parse_meta(response),
        )

    def get_all(
        self,
        *,
        name: str | None = None,
        mode: ZoneMode | None = None,
        label_selector: str | None = None,
        sort: list[str] | None = None,
    ) -> list[BoundZone]:
        """
        Returns a list of all Zone.

        See https://docs.hetzner.cloud/reference/cloud#zones-list-zones

        :param name: Filter resources by their name. The response will only contain the resources matching exactly the specified name.
        :param mode: Filter resources by their mode. The response will only contain the resources matching exactly the specified mode.
        :param label_selector: Filter resources by labels. The response will only contain resources matching the label selector.
        :param sort: Sort resources by field and direction.
        """
        return self._iter_pages(
            self.get_list,
            name=name,
            mode=mode,
            label_selector=label_selector,
            sort=sort,
        )

    def create(
        self,
        *,
        name: str,
        mode: ZoneMode,
        ttl: int | None = None,
        labels: dict[str, str] | None = None,
        primary_nameservers: list[ZonePrimaryNameserver] | None = None,
        rrsets: list[ZoneRRSet] | None = None,
        zonefile: str | None = None,
    ) -> CreateZoneResponse:
        """
        Creates a Zone.

        A default SOA and three NS resource records with the assigned Hetzner nameservers are created automatically.

        See https://docs.hetzner.cloud/reference/cloud#zones-create-a-zone

        :param name: Name of the Zone.
        :param mode: Mode of the Zone.
        :param ttl: Default Time To Live (TTL) of the Zone.
        :param labels: User-defined labels (key/value pairs) for the Resource.
        :param primary_nameservers: Primary nameservers of the Zone.
        :param rrsets: RRSets to be added to the Zone.
        :param zonefile: Zone file to import.
        """
        data: dict[str, Any] = {
            "name": name,
            "mode": mode,
        }
        if ttl is not None:
            data["ttl"] = ttl
        if labels is not None:
            data["labels"] = labels
        if primary_nameservers is not None:
            data["primary_nameservers"] = [o.to_payload() for o in primary_nameservers]
        if rrsets is not None:
            data["rrsets"] = [o.to_payload() for o in rrsets]
        if zonefile is not None:
            data["zonefile"] = zonefile

        response = self._client.request(
            method="POST",
            url=f"{self._base_url}",
            json=data,
        )

        return CreateZoneResponse(
            zone=BoundZone(self, response["zone"]),
            action=BoundAction(self._parent.actions, response["action"]),
        )

    def update(
        self,
        zone: Zone | BoundZone,
        *,
        labels: dict[str, str] | None = None,
    ) -> BoundZone:
        """
        Updates a Zone.

        See https://docs.hetzner.cloud/reference/cloud#zones-update-a-zone

        :param zone: Zone to update.
        :param labels: User-defined labels (key/value pairs) for the Resource.
        """
        data: dict[str, Any] = {}
        if labels is not None:
            data["labels"] = labels

        response = self._client.request(
            method="PUT",
            url=f"{self._base_url}/{zone.id_or_name}",
            json=data,
        )
        return BoundZone(self, response["zone"])

    def delete(
        self,
        zone: Zone | BoundZone,
    ) -> DeleteZoneResponse:
        """
        Deletes a Zone.

        See https://docs.hetzner.cloud/reference/cloud#zones-delete-a-zone

        :param zone: Zone to delete.
        """
        response = self._client.request(
            method="DELETE",
            url=f"{self._base_url}/{zone.id_or_name}",
        )

        return DeleteZoneResponse(
            action=BoundAction(self._parent.actions, response["action"]),
        )

    def export_zonefile(
        self,
        zone: Zone | BoundZone,
    ) -> ExportZonefileResponse:
        """
        Returns a generated Zone file in BIND (RFC 1034/1035) format.

        See https://docs.hetzner.cloud/reference/cloud#zones-export-a-zone-file

        :param zone: Zone to export the zone file from.
        """
        response = self._client.request(
            method="GET",
            url=f"{self._base_url}/{zone.id_or_name}/zonefile",
        )
        return ExportZonefileResponse(response["zonefile"])

    def get_actions_list(
        self,
        zone: Zone | BoundZone,
        *,
        status: list[ActionStatus] | None = None,
        sort: list[ActionSort] | None = None,
        page: int | None = None,
        per_page: int | None = None,
    ) -> ActionsPageResult:
        """
        Returns a paginated list of Actions for a Zone.

        See https://docs.hetzner.cloud/reference/cloud#zones-list-zones

        :param zone: Zone to get the Actions for.
        :param status: Filter the Actions by status.
        :param sort: Sort Actions by field and direction.
        :param page: Page number to get.
        :param per_page: Maximum number of Actions returned per page.
        """
        return self._get_actions_list(
            f"{self._base_url}/{zone.id_or_name}",
            status=status,
            sort=sort,
            page=page,
            per_page=per_page,
        )

    def get_actions(
        self,
        zone: Zone | BoundZone,
        *,
        status: list[ActionStatus] | None = None,
        sort: list[ActionSort] | None = None,
    ) -> list[BoundAction]:
        """
        Returns all Actions for a Zone.

        See https://docs.hetzner.cloud/reference/cloud#zones-list-zones

        :param zone: Zone to get the Actions for.
        :param status: Filter the Actions by status.
        :param sort: Sort Actions by field and direction.
        """
        return self._iter_pages(
            self.get_actions_list,
            zone,
            status=status,
            sort=sort,
        )

    def import_zonefile(
        self,
        zone: Zone | BoundZone,
        zonefile: str,
    ) -> BoundAction:
        """
        Imports a zone file, replacing all resource record sets (ZoneRRSet).

        See https://docs.hetzner.cloud/reference/cloud#zone-actions-import-a-zone-file

        :param zone: Zone to import the zone file into.
        :param zonefile: Zone file to import.
        """
        data: dict[str, Any] = {
            "zonefile": zonefile,
        }

        response = self._client.request(
            method="POST",
            url=f"{self._base_url}/{zone.id_or_name}/actions/import_zonefile",
            json=data,
        )
        return BoundAction(self._parent.actions, response["action"])

    def change_protection(
        self,
        zone: Zone | BoundZone,
        *,
        delete: bool | None = None,
    ) -> BoundAction:
        """
        Changes the protection of a Zone.

        See https://docs.hetzner.cloud/reference/cloud#zone-actions-change-a-zones-protection

        :param zone: Zone to update.
        :param delete: Prevents the Zone from being deleted.
        """
        data: dict[str, Any] = {}
        if delete is not None:
            data["delete"] = delete

        response = self._client.request(
            method="POST",
            url=f"{self._base_url}/{zone.id_or_name}/actions/change_protection",
            json=data,
        )
        return BoundAction(self._parent.actions, response["action"])

    def change_ttl(
        self,
        zone: Zone | BoundZone,
        ttl: int,
    ) -> BoundAction:
        """
        Changes the TTL of a Zone.

        See https://docs.hetzner.cloud/reference/cloud#zone-actions-change-a-zones-default-ttl

        :param zone: Zone to update.
        :param ttl: Default Time To Live (TTL) of the Zone.
        """
        data: dict[str, Any] = {
            "ttl": ttl,
        }

        response = self._client.request(
            method="POST",
            url=f"{self._base_url}/{zone.id_or_name}/actions/change_ttl",
            json=data,
        )
        return BoundAction(self._parent.actions, response["action"])

    def change_primary_nameservers(
        self,
        zone: Zone | BoundZone,
        primary_nameservers: list[ZonePrimaryNameserver],
    ) -> BoundAction:
        """
        Changes the primary nameservers of a Zone.

        See https://docs.hetzner.cloud/reference/cloud#zone-actions-change-a-zones-primary-nameservers

        :param zone: Zone to update.
        :param primary_nameservers: Primary nameservers of the Zone.
        """
        data: dict[str, Any] = {
            "primary_nameservers": [o.to_payload() for o in primary_nameservers],
        }

        response = self._client.request(
            method="POST",
            url=f"{self._base_url}/{zone.id_or_name}/actions/change_primary_nameservers",
            json=data,
        )
        return BoundAction(self._parent.actions, response["action"])

    def get_rrset(
        self,
        zone: Zone | BoundZone,
        name: str,
        type: ZoneRRSetType,
    ) -> BoundZoneRRSet:
        """
        Returns a single ZoneRRSet from the Zone.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-get-an-rrset

        :param zone: Zone to fetch the RRSet from.
        :param name: Name of the RRSet.
        :param type: Type of the RRSet.
        """
        response = self._client.request(
            method="GET",
            url=f"{self._base_url}/{zone.id_or_name}/rrsets/{name}/{type}",
        )
        return BoundZoneRRSet(self, response["rrset"])

    def get_rrset_list(
        self,
        zone: Zone | BoundZone,
        *,
        name: str | None = None,
        type: list[ZoneRRSetType] | None = None,
        label_selector: str | None = None,
        sort: list[str] | None = None,
        page: int | None = None,
        per_page: int | None = None,
    ) -> ZoneRRSetsPageResult:
        """
        Returns all ZoneRRSet in the Zone for a specific page.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-list-rrsets

        :param zone: Zone to fetch the RRSets from.
        :param name: Filter resources by their name. The response will only contain the resources matching exactly the specified name.
        :param type: Filter resources by their type. The response will only contain the resources matching exactly the specified type.
        :param label_selector: Filter resources by labels. The response will only contain resources matching the label selector.
        :param sort: Sort resources by field and direction.
        :param page: Page number to return.
        :param per_page: Maximum number of entries returned per page.
        """
        params: dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        if type is not None:
            params["type"] = type
        if label_selector is not None:
            params["label_selector"] = label_selector
        if sort is not None:
            params["sort"] = sort
        if page is not None:
            params["page"] = page
        if per_page is not None:
            params["per_page"] = per_page

        response = self._client.request(
            method="GET",
            url=f"{self._base_url}/{zone.id_or_name}/rrsets",
            params=params,
        )
        return ZoneRRSetsPageResult(
            rrsets=[BoundZoneRRSet(self, item) for item in response["rrsets"]],
            meta=Meta.parse_meta(response),
        )

    def get_rrset_all(
        self,
        zone: Zone | BoundZone,
        *,
        name: str | None = None,
        type: list[ZoneRRSetType] | None = None,
        label_selector: str | None = None,
        sort: list[str] | None = None,
    ) -> list[BoundZoneRRSet]:
        """
        Returns all ZoneRRSet in the Zone.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-list-rrsets

        :param zone: Zone to fetch the RRSets from.
        :param name: Filter resources by their name. The response will only contain the resources matching exactly the specified name.
        :param type: Filter resources by their type. The response will only contain the resources matching exactly the specified type.
        :param label_selector: Filter resources by labels. The response will only contain resources matching the label selector.
        :param sort: Sort resources by field and direction.
        """
        return self._iter_pages(
            self.get_rrset_list,
            zone,
            name=name,
            type=type,
            label_selector=label_selector,
            sort=sort,
        )

    def create_rrset(
        self,
        zone: Zone | BoundZone,
        *,
        name: str,
        type: ZoneRRSetType,
        ttl: int | None = None,
        labels: dict[str, str] | None = None,
        records: list[ZoneRecord] | None = None,
    ) -> CreateZoneRRSetResponse:
        """
        Creates a ZoneRRSet in the Zone.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-create-an-rrset

        :param zone: Zone to create the RRSets in.
        :param name: Name of the RRSet.
        :param type: Type of the RRSet.
        :param ttl: Time To Live (TTL) of the RRSet.
        :param labels: User-defined labels (key/value pairs) for the Resource.
        :param records: Records of the RRSet.
        """
        data: dict[str, Any] = {
            "name": name,
            "type": type,
        }
        if ttl is not None:
            data["ttl"] = ttl
        if labels is not None:
            data["labels"] = labels
        if records is not None:
            data["records"] = [o.to_payload() for o in records]

        response = self._client.request(
            method="POST",
            url=f"{self._base_url}/{zone.id_or_name}/rrsets",
            json=data,
        )
        return CreateZoneRRSetResponse(
            rrset=BoundZoneRRSet(self, response["rrset"]),
            action=BoundAction(self._parent.actions, response["action"]),
        )

    def update_rrset(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        *,
        labels: dict[str, str] | None = None,
    ) -> BoundZoneRRSet:
        """
        Updates a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-update-an-rrset

        :param rrset: RRSet to update.
        :param labels: User-defined labels (key/value pairs) for the Resource.
        """
        if rrset.zone is None:
            raise ValueError("rrset zone property is none")

        data: dict[str, Any] = {}
        if labels is not None:
            data["labels"] = labels

        response = self._client.request(
            method="PUT",
            url=f"{self._base_url}/{rrset.zone.id_or_name}/rrsets/{rrset.name}/{rrset.type}",
            json=data,
        )
        return BoundZoneRRSet(self, response["rrset"])

    def delete_rrset(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
    ) -> DeleteZoneRRSetResponse:
        """
        Deletes a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrsets-delete-an-rrset

        :param rrset: RRSet to delete.
        """
        if rrset.zone is None:
            raise ValueError("rrset zone property is none")

        response = self._client.request(
            method="DELETE",
            url=f"{self._base_url}/{rrset.zone.id_or_name}/rrsets/{rrset.name}/{rrset.type}",
        )
        return DeleteZoneRRSetResponse(
            action=BoundAction(self._parent.actions, response["action"]),
        )

    def change_rrset_protection(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        *,
        change: bool | None = None,
    ) -> BoundAction:
        """
        Changes the protection of a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-change-an-rrsets-protection

        :param rrset: RRSet to update.
        :param change: Prevent the Zone from being changed (deletion and updates).
        """
        if rrset.zone is None:
            raise ValueError("rrset zone property is none")

        data: dict[str, Any] = {}
        if change is not None:
            data["change"] = change

        response = self._client.request(
            method="POST",
            url=f"{self._base_url}/{rrset.zone.id_or_name}/rrsets/{rrset.name}/{rrset.type}/actions/change_protection",
            json=data,
        )
        return BoundAction(self._parent.actions, response["action"])

    def change_rrset_ttl(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        ttl: int | None,
    ) -> BoundAction:
        """
        Changes the TTL of a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-change-an-rrsets-ttl

        :param rrset: RRSet to update.
        :param change: Time To Live (TTL) of the RRSet.
        """
        if rrset.zone is None:
            raise ValueError("rrset zone property is none")

        data: dict[str, Any] = {
            "ttl": ttl,
        }

        response = self._client.request(
            method="POST",
            url=f"{self._base_url}/{rrset.zone.id_or_name}/rrsets/{rrset.name}/{rrset.type}/actions/change_ttl",
            json=data,
        )
        return BoundAction(self._parent.actions, response["action"])

    def add_rrset_records(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        records: list[ZoneRecord],
        ttl: int | None = None,
    ) -> BoundAction:
        """
        Adds records to a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-add-records-to-an-rrset

        :param rrset: RRSet to update.
        :param records: Records to add to the RRSet.
        :param ttl: Time To Live (TTL) of the RRSet.
        """
        if rrset.zone is None:
            raise ValueError("rrset zone property is none")

        data: dict[str, Any] = {
            "records": [o.to_payload() for o in records],
        }
        if ttl is not None:
            data["ttl"] = ttl

        response = self._client.request(
            method="POST",
            url=f"{self._base_url}/{rrset.zone.id_or_name}/rrsets/{rrset.name}/{rrset.type}/actions/add_records",
            json=data,
        )
        return BoundAction(self._parent.actions, response["action"])

    def update_rrset_records(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        records: list[ZoneRecord],
    ) -> BoundAction:
        """
        Updates records in a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-update-records-of-an-rrset

        :param rrset: RRSet to update.
        :param records: Records to update in the RRSet.
        """
        if rrset.zone is None:
            raise ValueError("rrset zone property is none")

        data: dict[str, Any] = {
            "records": [o.to_payload() for o in records],
        }

        response = self._client.request(
            method="POST",
            url=f"{self._base_url}/{rrset.zone.id_or_name}/rrsets/{rrset.name}/{rrset.type}/actions/update_records",
            json=data,
        )
        return BoundAction(self._parent.actions, response["action"])

    def remove_rrset_records(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        records: list[ZoneRecord],
    ) -> BoundAction:
        """
        Removes records from a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-remove-records-from-an-rrset

        :param rrset: RRSet to update.
        :param records: Records to remove from the RRSet.
        """
        if rrset.zone is None:
            raise ValueError("rrset zone property is none")

        data: dict[str, Any] = {
            "records": [o.to_payload() for o in records],
        }

        response = self._client.request(
            method="POST",
            url=f"{self._base_url}/{rrset.zone.id_or_name}/rrsets/{rrset.name}/{rrset.type}/actions/remove_records",
            json=data,
        )
        return BoundAction(self._parent.actions, response["action"])

    def set_rrset_records(
        self,
        rrset: ZoneRRSet | BoundZoneRRSet,
        records: list[ZoneRecord],
    ) -> BoundAction:
        """
        Sets the records of a ZoneRRSet.

        See https://docs.hetzner.cloud/reference/cloud#zone-rrset-actions-set-records-of-an-rrset

        :param rrset: RRSet to update.
        :param records: Records to set in the RRSet.
        """
        if rrset.zone is None:
            raise ValueError("rrset zone property is none")

        data: dict[str, Any] = {
            "records": [o.to_payload() for o in records],
        }

        response = self._client.request(
            method="POST",
            url=f"{self._base_url}/{rrset.zone.id_or_name}/rrsets/{rrset.name}/{rrset.type}/actions/set_records",
            json=data,
        )
        return BoundAction(self._parent.actions, response["action"])



from __future__ import annotations

from ._client import (
    Client,
    constant_backoff_function,
    exponential_backoff_function,
)
from ._exceptions import APIException, HCloudException
from ._version import __version__

__all__ = [
    "__version__",
    "Client",
    "constant_backoff_function",
    "exponential_backoff_function",
    "APIException",
    "HCloudException",
]



from __future__ import annotations

import time
import warnings
from typing import TYPE_CHECKING, Any, Literal, NamedTuple

from ..core import BoundModelBase, Meta, ResourceClientBase
from .domain import Action, ActionFailedException, ActionStatus, ActionTimeoutException

if TYPE_CHECKING:
    from .._client import Client


__all__ = [
    "ActionsClient",
    "ActionsPageResult",
    "BoundAction",
    "ResourceActionsClient",
    "ActionSort",
]


class BoundAction(BoundModelBase[Action], Action):
    _client: ActionsClient

    model = Action

    def wait_until_finished(self, max_retries: int | None = None) -> None:
        """Wait until the specific action has status=finished.

        :param max_retries: int Specify how many retries will be performed before an ActionTimeoutException will be raised.
        :raises: ActionFailedException when action is finished with status==error
        :raises: ActionTimeoutException when Action is still in status==running after max_retries is reached.
        """
        if max_retries is None:
            # pylint: disable=protected-access
            max_retries = self._client._client._poll_max_retries

        retries = 0
        while True:
            self.reload()
            if self.status != Action.STATUS_RUNNING:
                break

            retries += 1
            if retries < max_retries:
                # pylint: disable=protected-access
                time.sleep(self._client._client._poll_interval_func(retries))
                continue

            raise ActionTimeoutException(action=self)

        if self.status == Action.STATUS_ERROR:
            raise ActionFailedException(action=self)


ActionSort = Literal[
    "id",
    "id:asc",
    "id:desc",
    "command",
    "command:asc",
    "command:desc",
    "status",
    "status:asc",
    "status:desc",
    "started",
    "started:asc",
    "started:desc",
    "finished",
    "finished:asc",
    "finished:desc",
]


class ActionsPageResult(NamedTuple):
    actions: list[BoundAction]
    meta: Meta


class ResourceClientBaseActionsMixin(ResourceClientBase):
    def _get_action_by_id(
        self,
        base_url: str,
        id: int,
    ) -> BoundAction:
        response = self._client.request(
            method="GET",
            url=f"{base_url}/actions/{id}",
        )
        return BoundAction(
            client=self._parent.actions,
            data=response["action"],
        )

    def _get_actions_list(
        self,
        base_url: str,
        status: list[ActionStatus] | None = None,
        sort: list[ActionSort] | None = None,
        page: int | None = None,
        per_page: int | None = None,
    ) -> ActionsPageResult:
        params: dict[str, Any] = {}
        if status is not None:
            params["status"] = status
        if sort is not None:
            params["sort"] = sort
        if page is not None:
            params["page"] = page
        if per_page is not None:
            params["per_page"] = per_page

        response = self._client.request(
            method="GET",
            url=f"{base_url}/actions",
            params=params,
        )
        return ActionsPageResult(
            actions=[BoundAction(self._parent.actions, o) for o in response["actions"]],
            meta=Meta.parse_meta(response),
        )


class ResourceActionsClient(
    ResourceClientBaseActionsMixin,
    ResourceClientBase,
):
    _resource: str

    def __init__(self, client: ResourceClientBase | Client, resource: str | None):
        if isinstance(client, ResourceClientBase):
            super().__init__(client._parent)
            # Use the same base client as the the resource base client. Allows us to
            # choose the base client outside of the ResourceActionsClient.
            self._client = client._client
        else:
            # Backward compatibility, defaults to the parent ("top level") base client (`_client`).
            super().__init__(client)

        self._resource = resource or ""

    def get_by_id(self, id: int) -> BoundAction:
        """
        Returns a specific Action by its ID.

        :param id: ID of the Action.
        """
        return self._get_action_by_id(self._resource, id)

    def get_list(
        self,
        status: list[ActionStatus] | None = None,
        sort: list[ActionSort] | None = None,
        page: int | None = None,
        per_page: int | None = None,
    ) -> ActionsPageResult:
        """
        Returns a paginated list of Actions.

        :param status: Filter the Actions by status.
        :param sort: Sort Actions by field and direction.
        :param page: Page number to get.
        :param per_page: Maximum number of Actions returned per page.
        """
        return self._get_actions_list(
            self._resource,
            status=status,
            sort=sort,
            page=page,
            per_page=per_page,
        )

    def get_all(
        self,
        status: list[ActionStatus] | None = None,
        sort: list[ActionSort] | None = None,
    ) -> list[BoundAction]:
        """
        Returns all Actions.

        :param status: Filter the Actions by status.
        :param sort: Sort Actions by field and direction.
        """
        return self._iter_pages(self.get_list, status=status, sort=sort)


class ActionsClient(ResourceActionsClient):
    def __init__(self, client: Client):
        super().__init__(client, None)

    def get_list(
        self,
        status: list[ActionStatus] | None = None,
        sort: list[ActionSort] | None = None,
        page: int | None = None,
        per_page: int | None = None,
    ) -> ActionsPageResult:
        """
        .. deprecated:: 1.28
            Use :func:`client.<resource>.actions.get_list` instead,
            e.g. using :attr:`hcloud.certificates.client.CertificatesClient.actions`.

            `Starting 1 October 2023, it will no longer be available. <https://docs.hetzner.cloud/changelog#2023-07-20-actions-list-endpoint-is-deprecated>`_
        """
        warnings.warn(
            "The 'client.actions.get_list' method is deprecated, please use the "
            "'client.<resource>.actions.get_list' method instead (e.g. "
            "'client.certificates.actions.get_list').",
            DeprecationWarning,
            stacklevel=2,
        )
        return super().get_list(status=status, sort=sort, page=page, per_page=per_page)

    def get_all(
        self,
        status: list[ActionStatus] | None = None,
        sort: list[ActionSort] | None = None,
    ) -> list[BoundAction]:
        """
        .. deprecated:: 1.28
            Use :func:`client.<resource>.actions.get_all` instead,
            e.g. using :attr:`hcloud.certificates.client.CertificatesClient.actions`.

            `Starting 1 October 2023, it will no longer be available. <https://docs.hetzner.cloud/changelog#2023-07-20-actions-list-endpoint-is-deprecated>`_
        """
        warnings.warn(
            "The 'client.actions.get_all' method is deprecated, please use the "
            "'client.<resource>.actions.get_all' method instead (e.g. "
            "'client.certificates.actions.get_all').",
            DeprecationWarning,
            stacklevel=2,
        )
        return super().get_all(status=status, sort=sort)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, TypedDict

from ..core import BaseDomain, DomainIdentityMixin

if TYPE_CHECKING:
    from ..actions import BoundAction
    from .client import BoundZone, BoundZoneRRSet


__all__ = [
    "ZoneMode",
    "ZoneStatus",
    "ZoneRegistrar",
    "Zone",
    "ZonePrimaryNameserver",
    "ZoneAuthoritativeNameservers",
    "ZoneProtection",
    "CreateZoneResponse",
    "DeleteZoneResponse",
    "ExportZonefileResponse",
    "ZoneRRSet",
    "ZoneRRSetProtection",
    "ZoneRecord",
    "CreateZoneRRSetResponse",
    "DeleteZoneRRSetResponse",
]

ZoneMode = Literal["primary", "secondary"]
ZoneStatus = Literal["ok", "updating", "error"]
ZoneRegistrar = Literal["hetzner", "other", "unknown"]


class Zone(BaseDomain, DomainIdentityMixin):
    """
    Zone Domain.

    See https://docs.hetzner.cloud/reference/cloud#zones.
    """

    MODE_PRIMARY = "primary"
    """
    Zone in primary mode, resource record sets (RRSets) and resource records (RRs) are
    managed via the Cloud API or Cloud Console.
    """
    MODE_SECONDARY = "secondary"
    """
    Zone in secondary mode, Hetzner's nameservers query RRSets and RRs from given
    primary nameservers via AXFR.
    """

    STATUS_OK = "ok"
    """The Zone is pushed to the authoritative nameservers."""
    STATUS_UPDATING = "updating"
    """The Zone is currently being published to the authoritative nameservers."""
    STATUS_ERROR = "error"
    """The Zone could not be published to the authoritative nameservers."""

    REGISTRAR_HETZNER = "hetzner"
    REGISTRAR_OTHER = "other"
    REGISTRAR_UNKNOWN = "unknown"

    __api_properties__ = (
        "id",
        "name",
        "created",
        "mode",
        "ttl",
        "labels",
        "protection",
        "status",
        "record_count",
        "registrar",
        "primary_nameservers",
        "authoritative_nameservers",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        id: int | None = None,
        name: str | None = None,
        created: str | None = None,
        mode: ZoneMode | None = None,
        ttl: int | None = None,
        labels: dict[str, str] | None = None,
        protection: ZoneProtection | None = None,
        status: ZoneStatus | None = None,
        record_count: int | None = None,
        registrar: ZoneRegistrar | None = None,
        primary_nameservers: list[ZonePrimaryNameserver] | None = None,
        authoritative_nameservers: ZoneAuthoritativeNameservers | None = None,
    ):
        self.id = id
        self.name = name
        self.created = self._parse_datetime(created)
        self.mode = mode
        self.ttl = ttl
        self.labels = labels
        self.protection = protection
        self.status = status
        self.record_count = record_count
        self.registrar = registrar
        self.primary_nameservers = primary_nameservers
        self.authoritative_nameservers = authoritative_nameservers


ZonePrimaryNameserverTSIGAlgorithm = Literal[
    "hmac-md5",
    "hmac-sha1",
    "hmac-sha256",
]


class ZonePrimaryNameserver(BaseDomain):
    """
    Zone Primary Nameserver Domain.
    """

    TSIG_ALGORITHM_HMAC_MD5 = "hmac-md5"
    """Transaction signature (TSIG) algorithm used to generate the TSIG key."""
    TSIG_ALGORITHM_HMAC_SHA1 = "hmac-sha1"
    """Transaction signature (TSIG) algorithm used to generate the TSIG key."""
    TSIG_ALGORITHM_HMAC_SHA256 = "hmac-sha256"
    """Transaction signature (TSIG) algorithm used to generate the TSIG key."""

    __api_properties__ = (
        "address",
        "port",
        "tsig_algorithm",
        "tsig_key",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        address: str,
        port: int | None = None,
        tsig_algorithm: ZonePrimaryNameserverTSIGAlgorithm | None = None,
        tsig_key: str | None = None,
    ):
        self.address = address
        self.port = port
        self.tsig_algorithm = tsig_algorithm
        self.tsig_key = tsig_key

    def to_payload(self) -> dict[str, Any]:
        """
        Generates the request payload from this domain object.
        """
        payload: dict[str, Any] = {
            "address": self.address,
        }
        if self.port is not None:
            payload["port"] = self.port
        if self.tsig_algorithm is not None:
            payload["tsig_algorithm"] = self.tsig_algorithm
        if self.tsig_key is not None:
            payload["tsig_key"] = self.tsig_key

        return payload


ZoneAuthoritativeNameserversDelegationStatus = Literal[
    "valid",
    "partially-valid",
    "invalid",
    "lame",
    "unregistered",
    "unknown",
]


class ZoneAuthoritativeNameservers(BaseDomain):
    """
    Zone Authoritative Nameservers Domain.
    """

    DELEGATION_STATUS_VALID = "valid"
    DELEGATION_STATUS_PARTIALLY_VALID = "partially-valid"
    DELEGATION_STATUS_INVALID = "invalid"
    DELEGATION_STATUS_LAME = "lame"
    DELEGATION_STATUS_UNREGISTERED = "unregistered"
    DELEGATION_STATUS_UNKNOWN = "unknown"

    __api_properties__ = (
        "assigned",
        "delegated",
        "delegation_last_check",
        "delegation_status",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        assigned: list[str] | None = None,
        delegated: list[str] | None = None,
        delegation_last_check: str | None = None,
        delegation_status: ZoneAuthoritativeNameserversDelegationStatus | None = None,
    ):
        self.assigned = assigned
        self.delegated = delegated
        self.delegation_last_check = self._parse_datetime(delegation_last_check)
        self.delegation_status = delegation_status


class ZoneProtection(TypedDict):
    """
    Zone Protection.
    """

    delete: bool


class CreateZoneResponse(BaseDomain):
    """
    Create Zone Response Domain.
    """

    __api_properties__ = ("zone", "action")
    __slots__ = __api_properties__

    def __init__(
        self,
        zone: BoundZone,
        action: BoundAction,
    ):
        self.zone = zone
        self.action = action


class DeleteZoneResponse(BaseDomain):
    """
    Delete Zone Response Domain.
    """

    __api_properties__ = ("action",)
    __slots__ = __api_properties__

    def __init__(
        self,
        action: BoundAction,
    ):
        self.action = action


class ExportZonefileResponse(BaseDomain):
    """
    Export Zonefile Response Domain.
    """

    __api_properties__ = ("zonefile",)
    __slots__ = __api_properties__

    def __init__(
        self,
        zonefile: str,
    ):
        self.zonefile = zonefile


ZoneRRSetType = Literal[
    "A",
    "AAAA",
    "CAA",
    "CNAME",
    "DS",
    "HINFO",
    "HTTPS",
    "MX",
    "NS",
    "PTR",
    "RP",
    "SOA",
    "SRV",
    "SVCB",
    "TLSA",
    "TXT",
]


class ZoneRRSet(BaseDomain):
    """
    Zone RRSet Domain.

    See https://docs.hetzner.cloud/reference/cloud#zone-rrsets
    """

    TYPE_A = "A"
    TYPE_AAAA = "AAAA"
    TYPE_CAA = "CAA"
    TYPE_CNAME = "CNAME"
    TYPE_DS = "DS"
    TYPE_HINFO = "HINFO"
    TYPE_HTTPS = "HTTPS"
    TYPE_MX = "MX"
    TYPE_NS = "NS"
    TYPE_PTR = "PTR"
    TYPE_RP = "RP"
    TYPE_SOA = "SOA"
    TYPE_SRV = "SRV"
    TYPE_SVCB = "SVCB"
    TYPE_TLSA = "TLSA"
    TYPE_TXT = "TXT"

    __api_properties__ = (
        "name",
        "type",
        "ttl",
        "labels",
        "protection",
        "records",
        "id",
        "zone",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        name: str | None = None,
        type: ZoneRRSetType | None = None,
        ttl: int | None = None,
        labels: dict[str, str] | None = None,
        protection: ZoneRRSetProtection | None = None,
        records: list[ZoneRecord] | None = None,
        id: str | None = None,
        zone: BoundZone | Zone | None = None,
    ):
        # Ensure that 'id', 'name' and 'type' are always populated.
        if name is not None and type is not None:
            if id is None:
                id = f"{name}/{type}"
        else:
            if id is not None:
                name, _, type = id.partition("/")  # type: ignore[assignment]
            else:
                raise ValueError("id or name and type must be set")

        self.name = name
        self.type = type
        self.ttl = ttl
        self.labels = labels
        self.protection = protection
        self.records = records

        self.id = id
        self.zone = zone

    def to_payload(self) -> dict[str, Any]:
        """
        Generates the request payload from this domain object.
        """
        payload: dict[str, Any] = {
            "name": self.name,
            "type": self.type,
        }
        if self.ttl is not None:
            payload["ttl"] = self.ttl
        if self.labels is not None:
            payload["labels"] = self.labels
        if self.protection is not None:
            payload["protection"] = self.protection
        if self.records is not None:
            payload["records"] = [o.to_payload() for o in self.records]

        return payload


class ZoneRRSetProtection(TypedDict):
    """
    Zone RRSet Protection.
    """

    change: bool


class ZoneRecord(BaseDomain):
    """
    Zone Record Domain.
    """

    __api_properties__ = (
        "value",
        "comment",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        value: str,
        comment: str | None = None,
    ):
        self.value = value
        self.comment = comment

    def to_payload(self) -> dict[str, Any]:
        """
        Generates the request payload from this domain object.
        """
        payload: dict[str, Any] = {
            "value": self.value,
        }
        if self.comment is not None:
            payload["comment"] = self.comment

        return payload


class CreateZoneRRSetResponse(BaseDomain):
    """
    Create Zone RRSet Response Domain.
    """

    __api_properties__ = (
        "rrset",
        "action",
    )
    __slots__ = __api_properties__

    def __init__(
        self,
        rrset: BoundZoneRRSet,
        action: BoundAction,
    ):
        self.rrset = rrset
        self.action = action


class DeleteZoneRRSetResponse(BaseDomain):
    """
    Delete Zone RRSet Response Domain.
    """

    __api_properties__ = ("action",)
    __slots__ = __api_properties__

    def __init__(
        self,
        action: BoundAction,
    ):
        self.action = action



from __future__ import annotations

from typing import TYPE_CHECKING, Any, NamedTuple

from ..core import BoundModelBase, Meta, ResourceClientBase
from .domain import StorageBoxType

if TYPE_CHECKING:
    from .._client import Client

__all__ = [
    "BoundStorageBoxType",
    "StorageBoxTypesPageResult",
    "StorageBoxTypesClient",
]


class BoundStorageBoxType(BoundModelBase[StorageBoxType], StorageBoxType):
    _client: StorageBoxTypesClient

    model = StorageBoxType


class StorageBoxTypesPageResult(NamedTuple):
    storage_box_types: list[BoundStorageBoxType]
    meta: Meta


class StorageBoxTypesClient(ResourceClientBase):
    """
    A client for the Storage Box Types API.

    See https://docs.hetzner.cloud/reference/hetzner#storage-box-types.
    """

    _base_url = "/storage_box_types"

    def __init__(self, client: Client):
        super().__init__(client)
        self._client = client._client_hetzner

    def get_by_id(self, id: int) -> BoundStorageBoxType:
        """
        Returns a specific Storage Box Type.

        See https://docs.hetzner.cloud/reference/hetzner#storage-box-types-get-a-storage-box-type

        :param id: ID of the Storage Box Type.
        """
        response = self._client.request(
            method="GET",
            url=f"{self._base_url}/{id}",
        )
        return BoundStorageBoxType(self, response["storage_box_type"])

    def get_by_name(self, name: str) -> BoundStorageBoxType | None:
        """
        Returns a specific Storage Box Type.

        See https://docs.hetzner.cloud/reference/hetzner#storage-box-types-list-storage-box-types

        :param name: Name of the Storage Box Type.
        """
        return self._get_first_by(self.get_list, name=name)

    def get_list(
        self,
        name: str | None = None,
        page: int | None = None,
        per_page: int | None = None,
    ) -> StorageBoxTypesPageResult:
        """
        Returns a list of Storage Box Types for a specific page.

        See https://docs.hetzner.cloud/reference/hetzner#storage-box-types-list-storage-box-types

        :param name: Name of the Storage Box Type.
        :param page: Page number to return.
        :param per_page: Maximum number of entries returned per page.
        """
        params: dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        if page is not None:
            params["page"] = page
        if per_page is not None:
            params["per_page"] = per_page

        response = self._client.request(
            method="GET",
            url=f"{self._base_url}",
            params=params,
        )
        return StorageBoxTypesPageResult(
            storage_box_types=[
                BoundStorageBoxType(self, o) for o in response["storage_box_types"]
            ],
            meta=Meta.parse_meta(response),
        )

    def get_all(
        self,
        name: str | None = None,
    ) -> list[BoundStorageBoxType]:
        """
        Returns all Storage Box Types.

        See https://docs.hetzner.cloud/reference/hetzner#storage-box-types-list-storage-box-types

        :param name: Name of the Storage Box Type.
        """
        return self._iter_pages(
            self.get_list,
            name=name,
        )
